from functools import wraps

import requests

from jovian.utils.credentials import get_api_key, purge_api_key
from jovian.utils.logger import log


def retry(request):
    """Decorator to make get/post requests retryable"""
    @wraps(request)
    def _request_wrapper(*args, **kwargs):
        for i in range(2):
            res = request(*args, **kwargs)
            if res.status_code == 401:
                log('The current API key is invalid or expired.', error=True)
                purge_api_key()

                # This will ensure that fresh api token is requested
                if 'headers' in kwargs:
                    kwargs['headers']['Authorization'] = "Bearer " + get_api_key()
            else:
                return res
        return res

    return _request_wrapper


@retry
def get(url, params=None, **kwargs):
    """Retryable GET request"""
    return requests.get(url, params=params, **kwargs)


@retry
def post(url, data=None, json=None, **kwargs):
    """Retryable POST request"""
    return requests.post(url, data=data, json=json, **kwargs)


def _msg(res):
    try:
        data = res.json()
        if 'errors' in data and len(data['errors']) > 0:
            return data['errors'][0]['message']
        if 'message' in data:
            return data['message']
        if 'msg' in data:
            return data['msg']
    except:
        if res.text:
            return res.text
        return 'Something went wrong'


def pretty(res):
    """Make a human readable output from an HTML response"""
    return '(HTTP ' + str(res.status_code) + ') ' + _msg(res)



import os
import click
from requests import get

from jovian._version import __version__
from jovian.utils.constants import ISSUES_MSG

from jovian.utils.credentials import get_guest_key, read_api_key_opt, read_api_url, read_org_id
from jovian.utils.logger import log
from jovian.utils.rcfile import get_rcdata, rcfile_exists, set_notebook_slug
from jovian.utils.request import pretty
from jovian.utils.shared import _u, _v


def _h(fresh):
    """Create a header to provide library metadata"""
    api_key, _ = read_api_key_opt()

    headers = {"x-jovian-source": "library",
               "x-jovian-library-version": __version__,
               "x-jovian-command": "clone" if fresh else "pull",
               "x-jovian-guest": get_guest_key(),
               "x-jovian-org": read_org_id()}

    if api_key is not None:
        headers["Authorization"] = "Bearer " + api_key

    return headers


def get_gist(slug, version, fresh):
    """Download a gist"""
    if '/' in slug:
        parts = slug.split('/')
        username, title = parts[0], parts[1]
        url = _u('user/' + username + '/gist/' + title + _v(version))
    else:
        url = _u('gist/' + slug + _v(version))
    res = get(url, headers=_h(fresh))
    if res.status_code == 200:
        return res.json()['data']
    raise Exception('Failed to retrieve Gist: ' + pretty(res))


def post_clone_msg(title):

    log("Cloned successfully to '{}'".format(title), color='green')
    log(click.style('\nNext steps:', fg='yellow', underline=True) +
        click.style("""
  $ cd {}
  $ jovian install
  $ conda activate <env_name>
  $ jupyter notebook
""".format(title), bold=True), pre=False)

    log("""
Replace <env_name> with the name of your environment (without the '<' & '>')
Jovian uses Anaconda ( https://conda.io/ ) under the hood,
so please make sure you have it installed and added to path.
* If you face issues with `jovian install`, try `conda env update`.
* If you face issues with `conda activate`, try `source activate <env_name>`
  or `activate <env_name>` to activate the virtual environment.
""", pre=False)


def clone(slug, version=None, fresh=True, include_outputs=True, overwrite=False):
    """Download the files for a gist"""
    # Print issues link
    log(ISSUES_MSG)

    # Download gist metadata
    ver_str = '(version ' + str(version) + ')' if version else ''
    log('Fetching ' + slug + " " + ver_str + "..")
    gist = get_gist(slug, version, fresh)
    title = gist['title']

    # If fresh clone, create directory
    if fresh and not os.path.exists(title):
        os.makedirs(title)
        os.chdir(title)

    elif fresh and os.path.exists(title) and overwrite:
        os.chdir(title)

    elif fresh and os.path.exists(title) and not overwrite:
        i = 1
        while os.path.exists(title + '-' + str(i)):
            i += 1
        title = title + '-' + str(i)

        os.makedirs(title)
        os.chdir(title)

    # Download the files
    log('Downloading files..')
    for f in gist['files']:
        if not f['artifact'] or include_outputs:
            with open(f['filename'], 'wb') as fp:
                fp.write(get(f['rawUrl']).content)

        # Create .jovianrc for a fresh clone
        if fresh and f['filename'].endswith('.ipynb'):
            set_notebook_slug(f['filename'], slug)

    # Print success message and instructions
    if fresh:
        post_clone_msg(title)
    else:
        log('Files dowloaded successfully in current directory')


RCFILE_NOTFOUND = "Could not detect '.jovianrc' file. Make sure you are running 'jovian pull' inside a directory cloned with 'jovian clone'."


def pull(slug=None, version=None):
    """Get the latest files associated with the current gist"""
    # If a slug is provided, just use that
    if slug:
        clone(slug, version, fresh=False)
        return

    # Check if .jovianrc exists
    if not rcfile_exists():
        log(RCFILE_NOTFOUND, error=True)
        return

    # Get list of notebooks
    nbs = get_rcdata()['notebooks']

    for fname in nbs:
        # Get the latest files for each notebook
        clone(nbs[fname]['slug'], version, fresh=False)



"""Git related utilities"""
import os
import subprocess


def is_git():
    """Check whether we are in a git repository"""
    task = subprocess.Popen('git rev-parse', shell=True, stderr=subprocess.PIPE)
    task.communicate()
    return task.returncode == 0


def get_branch():
    return os.popen('git rev-parse --abbrev-ref HEAD').read().strip()


def get_remote():
    """Get the remote URL"""
    return os.popen('git config --get remote.origin.url').read().replace(".git\n", "").strip()


def get_current_commit():
    """Get the current commit"""
    return os.popen('git rev-parse HEAD').read().strip()


def get_repository_root():
    """Get the root of the git repository"""
    return os.popen('git rev-parse --show-toplevel').read().strip()


def commit(message):
    """Create a new commit"""
    return os.system('cd ' + get_repository_root() + ' && git add . &&  git commit -a -m "' + message + '" && cd -')


def git_push():
    """Push the branch to origin"""
    return os.system("git push origin " + get_branch())


def git_commit_push(message):
    """Create a commit and push to origin"""
    if is_git():
        commit(message)
        git_push()
        return {
            'remote': get_remote(),
            'branch': get_branch(),
            'commit': get_current_commit()
        }


def get_relative_path():
    """Returns relative path from the git root directory."""
    root_dir = get_repository_root()
    file_dir = os.path.abspath('')
    return os.path.relpath(file_dir, root_dir)



from jovian._version import __version__
from jovian.utils.credentials import get_api_key, get_guest_key, read_api_url, read_org_id
from jovian.utils.error import ApiError
from jovian.utils.logger import log
from jovian.utils.request import get, pretty
from jovian.utils.shared import _u, _v
from jovian.utils.api import post_slack_message


def _h():
    """Create a header to provide library metadata"""
    return {"Authorization": "Bearer " + get_api_key(),
            "x-jovian-source": "library",
            "x-jovian-library-version": __version__,
            "x-jovian-command": "add-slack",
            "x-jovian-guest": get_guest_key(),
            "x-jovian-org": read_org_id()}


def add_slack():
    """prints instructions for connecting Slack, if Slack connection is not already present.
    if Slack is already connected, prints details about the workspace and the channel"""
    url = _u('/slack/integration_details')
    res = get(url, headers=_h())
    if res.status_code == 200:
        res = res.json()
        if not res.get('errors'):
            slack_account = res.get('data').get('slackAccount')
            log('Slack already connected. \nWorkspace: {}\nConnected Channel: {}'
                .format(slack_account.get('workspace'), slack_account.get('channel')))
        else:
            log(str(res.get('errors')[0].get('message')))
    else:
        raise ApiError('Slack trigger failed: ' + pretty(res))


def notify(data, verbose=True, safe=False):
    """Sends the data to the `Slack`_ workspace connected with your `Jovian`_ account.

    Args:
        data(dict|string): A dict or string to be pushed to Slack

        verbose(bool, optional): By default it prints the acknowledgement, you can remove this by setting the argument to False.

        safe(bool, optional): To avoid raising ApiError exception. Defaults to False.

    Example
        .. code-block::

            import jovian

            data = "Hello from the Integration!"
            jovian.notify(data)

    .. important::
        This feature requires for your Jovian account to be connected to a Slack workspace, visit `Jovian Integrations`_ to integrate them and to control the type of notifications.
    .. _Slack: https://slack.com
    .. _Jovian: https://jovian.ml?utm_source=docs
    .. _Jovian Integrations: https://jovian.ml/settings/integrations?utm_source=docs
    """
    res = post_slack_message(data=data, safe=safe)
    if verbose:
        if not res.get('errors'):
            log('message_sent:' + str(res.get('data').get('messageSent')))
        else:
            log(str(res.get('errors')[0].get('message')), error=True)



from jovian.utils import api
from jovian.utils.logger import log

_data_blocks = []


def get_records(slug_only=False):
    """Get the list of slugs recorded so far"""
    global _data_blocks
    if slug_only:
        return [slug for slug, _, _ in _data_blocks]
    else:
        return _data_blocks


def reset(*record_types):
    """Reset the tracked hyperparameters, metrics or dataset (for a fresh experiment)

    Args:
        *record_types(strings, optional): By default, resets all type of records. 
            To reset specific type of records, pass arguments `metrics`, 
            `hyperparams`, `dataset`
    Example
        .. code-block::

            import jovian
            jovian.reset('hyperparams', 'metrics')
    """
    global _data_blocks

    if len(record_types) == 0:
        _data_blocks = []
    else:
        _data_blocks = [(slug, rec_type, data) for (slug, rec_type, data) in _data_blocks
                        if rec_type not in record_types]


def _parse_data(data, data_args):
    """Parse different types of arguments"""
    data = data or {}
    if isinstance(data, dict):
        for k in data_args:
            data[k] = data_args[k]
    elif isinstance(data, list) and len(data_args.keys()) > 0:
        data.append(data_args)
    return data if len(data) > 0 else None


def log_record(record_type, data=None, verbose=True, **data_args):
    """Create records with the given data & type"""
    global _data_blocks
    # Create the combined data dictionary
    data = _parse_data(data, data_args)
    if data is None and verbose:
        log('Nothing to record. Skipping..', error=True)
        return
    # Send to API endpoint
    res = api.post_block(data, record_type)
    tracking_slug = res['tracking']['trackingSlug']
    # Save to data block
    _data_blocks.append((tracking_slug, record_type, data))
    if verbose:
        log(record_type.capitalize() + ' logged.')


def log_hyperparams(data_dict=None, verbose=True, **data_args):
    """Record hyperparameters for the current experiment

    Args:
        data_dict(dict, optional): A python dict to be recorded as hyperparmeters.

        verbose(bool, optional): By default it prints the acknowledgement, you can remove 
            this by setting the argument to False.

        **data_args(optional): Instead of passing a dictionary, you can also pass each 
            individual key-value pair as a argument (see example below)

    Example
        .. code-block::

            import jovian
            jovian.log_hyperparams(arch='cnn', lr=0.001)
            # or
            jovian.log_hyperparams({ 'arch': 'cnn', 'lr': 0.001 })
    """
    log_record('hyperparams', data_dict, verbose, **data_args)


def log_metrics(data_dict=None, verbose=True, **data_args):
    """Record metrics for the current experiment

    Args:
        data_dict(dict, optional): A python dict to be recorded as metrics.

        verbose(bool, optional): By default it prints the acknowledgement, you can remove 
            this by setting the argument to False.

        **data_args(any, optional): Instead of passing a dictionary, you can also pass each 
            individual key-value pair as a argument (see example below)

    Example
        .. code-block::

            import jovian
            jovian.log_metrics(epochs=1, train_loss=0.5, 
                               val_loss=0.3, val_accuracy=0.9)
            # or
            jovian.log_metrics({ 'epochs': 1, 'train_loss': 0.5 })
    """
    log_record('metrics', data_dict, verbose, **data_args)


def log_dataset(data_dict=None, verbose=True, **data_args):
    """Record dataset details for the current experiment

    Args:
        data_dict(dict, optional): A python dict to be recorded as dataset.

        verbose(bool, optional): By default it prints the acknowledgement, you can remove 
            this by setting the argument to False.

        **data_args(optional): Instead of passing a dictionary, you can also pass each 
            individual key-value pair as a argument (see example below)

    Example
        .. code-block::

            import jovian

            path = 'data/mnist'
            description = '28x28 images of handwritten digits (in grayscale)'

            jovian.log_dataset(path=path, description=description)
            # or 
            jovian.log_dataset({ 'path': path, 'description': description })
    """
    log_record('dataset', data_dict, verbose, **data_args)


def log_git(data_dict=None, verbose=True, **data_args):
    """Record the git-related information.

    Args:
        data(dict): A python dict or a array of dicts to be recorded as a git related block.

        verbose(bool, optional): By default it prints the acknowledgement, you can 
            remove this by setting the argument to False.
    """
    log_record('git', data_dict, verbose, **data_args)



import os


def setup_extension(enable=True):
    if enable:
        os.system("jupyter nbextension enable jovian_nb_ext/main --sys-prefix")
    else:
        os.system("jupyter nbextension disable jovian_nb_ext/main --sys-prefix")



"""Utilities to read, write and manage the credentials file"""
import json
import os
import shutil
import stat
import uuid
from uuid import UUID

import click
import requests

from jovian.utils.constants import DEFAULT_API_URL, DEFAULT_ORG_ID, DEFAULT_WEBAPP_URL
from jovian.utils.error import ApiError, ConfigError
from jovian.utils.logger import log
from jovian.utils.misc import is_flavor_pro, urljoin

try:
    # Python 3
    from json.decoder import JSONDecodeError
except ImportError:  # no-cover
    # Python 2
    JSONDecodeError = ValueError


# Keys used in credentials file
API_TOKEN_KEY = "API_KEY"
GUEST_TOKEN_KEY = "GUEST_KEY"
ORG_ID_KEY = "ORG_ID"
API_URL_KEY = "API_URL"
WEBAPP_URL_KEY = "WEBAPP_URL"

# Paths and filenames
HOME = os.path.expanduser('~')
CONFIG_DIR = HOME + '/.jovian'
CREDS_FNAME = 'credentials.json'
CONTACT_MSG = 'Looks like there\'s something wrong with your setup. Please report this issue to hello@jovian.ml'

# Config directory management


def get_creds_path():
    return os.path.join(CONFIG_DIR, CREDS_FNAME)


def config_exists():
    """Check if config directory exists"""
    return os.path.exists(CONFIG_DIR)


def purge_config():
    """Remove the config directory"""
    return shutil.rmtree(CONFIG_DIR, ignore_errors=True)


def init_config():
    """Create the config directory"""
    if not os.path.exists(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)


# Credentials file management

def purge_creds():
    """Remove the credentials file"""
    creds_path = get_creds_path()
    if os.path.isfile(creds_path):
        os.remove(creds_path)


def read_creds():
    """Read the credentials file"""
    creds_path = get_creds_path()
    if not os.path.exists(creds_path):
        return {}
    with open(creds_path, 'r') as f:
        try:
            return json.load(f)
        except ValueError:
            purge_creds()
            return {}


def creds_exist():
    """Check if credentials file exits"""
    creds = read_creds()
    return API_TOKEN_KEY in creds


def read_cred(key, default=None):
    """Read a particular key from the credentials file"""
    creds = read_creds()
    if default:
        return creds.get(key, default)
    return creds[key]


def write_creds(creds, update_cache=True):
    """Write the given credentials to file"""
    init_config()
    creds_path = get_creds_path()
    with open(creds_path, 'w') as f:
        json.dump(creds, f)
    os.chmod(creds_path, stat.S_IREAD | stat.S_IWRITE)


def write_cred(key, value):
    creds = read_creds()
    if key in creds and creds[key] == value:
        return
    creds[key] = value
    write_creds(creds)


def purge_cred_key(key):
    """Remove a particular key from config"""
    creds = read_creds()
    if key in creds:
        del creds[key]
        write_creds(creds)


# API URL

def write_api_url(value):
    """Write the API URL"""
    write_cred(API_URL_KEY, value)


def read_api_url():
    """Read the API URL"""
    ensure_org()
    return read_cred(API_URL_KEY, DEFAULT_API_URL)


# Webapp URL

def write_webapp_url(value):
    """Write the webapp URL"""
    write_cred(WEBAPP_URL_KEY, value)


def read_webapp_url():
    """Read the webapp URL"""
    ensure_org()
    return read_cred(WEBAPP_URL_KEY, DEFAULT_WEBAPP_URL)


# Organization ID

def write_org_id(value):
    """Write the Organization ID"""
    write_cred(ORG_ID_KEY, value)


def request_org_id():
    """Ask the user to provide the organization ID"""
    log("If you're a jovian-pro user please enter your company's organization ID on Jovian (otherwise leave it blank).")

    msg = "Organization ID"
    return click.prompt(msg, default='', show_default=False)


def read_org_id():
    """Read Organization ID"""
    ensure_org()
    return read_cred(ORG_ID_KEY, DEFAULT_ORG_ID)


def ensure_org(check_pro=True):
    """Check and set Organization ID"""
    # Check the flavor
    if check_pro and not is_flavor_pro():
        return

    # Read the credentials
    creds = read_creds()
    try:
        org_id = creds[ORG_ID_KEY]
        api_url = creds[API_URL_KEY]
        webapp_url = creds[WEBAPP_URL_KEY]
        if org_id and api_url and webapp_url:
            return
    except KeyError:
        pass

    # Request organization
    org_id = request_org_id()

    # Construct the webapp URL
    if org_id:
        webapp_url = 'https://' + org_id + '.jovian.ml/'
    else:
        org_id = DEFAULT_ORG_ID
        webapp_url = 'https://jovian.ml/'

    # Try to retrieve the config.json file from webapp
    try:
        config_url = webapp_url + 'config.json'
        config_res = requests.get(config_url)
    except ConnectionError as e:
        msg = 'Failed to connect to ' + webapp_url + \
            ' . Please verify your organization ID and ensure you are connected to the internet.'
        log(msg, error=True)
        raise ConfigError(msg, e)

    # Check for a successful response
    if config_res.status_code != 200:
        msg = 'Request to retrieve configuration file ' + config_url + \
            ' failed with status_code ' + str(config_res.status_code) + ' . ' + CONTACT_MSG
        log(msg, error=True)
        raise ConfigError(msg + ' Response (truncated):\n' +
                          config_res.text[:100])

    # Parse JSON configuration
    try:
        config_json = config_res.json()
    except JSONDecodeError as e:
        msg = 'Failed to parse JSON configuration file from ' + \
            config_url + ' . ' + CONTACT_MSG
        log(msg, error=True)
        raise ConfigError(msg + ' Response (truncated):\n' +
                          config_res.text[:100], e)

    # Extract API URL
    try:
        api_url = config_json[API_URL_KEY]
    except KeyError as e:
        msg = 'Failed to extract API_URL from JSON configuration file ' + \
            config_url + ' . ' + CONTACT_MSG
        log(msg, error=True)
        raise ConfigError(msg, e)

    # Save details to credentials file
    write_org_id(org_id)
    write_api_url(api_url)
    write_webapp_url(webapp_url)


def purge_api_key():
    """Remove API token from config"""
    purge_cred_key(API_TOKEN_KEY)


def write_api_key(value):
    """Write the API key to memory, and the credentials file"""
    write_cred(API_TOKEN_KEY, value)


def _u(path):
    """Make a URL from the path"""
    return urljoin(read_api_url(), path)


def validate_api_key(key):
    """Validate the API key by making a request to server"""
    res = requests.get(_u('/user/profile'), headers={'Authorization': 'Bearer ' + key})
    return res.status_code == 200


def get_api_key():
    """Retrieve and validate the API Key (from memory, config or user input)"""
    creds = read_creds()
    if API_TOKEN_KEY not in creds:
        key, _ = read_or_request_api_key()
        if not validate_api_key(key):
            log('The current API key is invalid or expired.', error=True)
            key, _ = request_api_key(), 'request'
            if not validate_api_key(key):
                raise ApiError('The API key provided is invalid or expired.')
        write_api_key(key)
        return key
    return creds[API_TOKEN_KEY]


# Guest token

def write_guest_key(token):
    """Write the GUEST key to memory, and the credentials file"""
    write_cred(GUEST_TOKEN_KEY, token)


def request_api_key():
    """Ask the user to provide the API key"""
    log("Please enter your API key ( from " + read_webapp_url() + " ):")
    api_key = click.prompt("API KEY", hide_input=True)
    return api_key


def read_api_key_opt():
    """Read credentials file, or return None"""
    creds = read_creds()
    return creds.get(API_TOKEN_KEY), 'read'


def read_or_request_api_key():
    """Read credentials file, or ask the user for API Key, if required"""
    api_key, source = read_api_key_opt()

    if api_key is not None:
        return api_key, source
    else:
        return request_api_key(), 'request'


def _generate_guest_key():
    """Generate GUEST key"""
    return uuid.uuid4().hex


def _read_or_generate_guest_key():
    """Read credentials file, or generate GUEST key, if required"""
    creds = read_creds()
    return creds.get(GUEST_TOKEN_KEY, _generate_guest_key())


def _validate_guest_key(key):
    """Validate GUEST key"""
    try:
        val = UUID(key, version=4)
    except ValueError:
        return False
    return val.hex == key


def get_guest_key():
    """Retrieve or generate the guest key"""
    # Read credentials
    creds = read_creds()
    guest_key = creds.get(GUEST_TOKEN_KEY, '')
    # Check validity & generate key (if required)
    if not _validate_guest_key(guest_key):
        write_cred(GUEST_TOKEN_KEY, _generate_guest_key())
        creds = read_creds()
    # Return final key
    return creds[GUEST_TOKEN_KEY]



from jovian.utils.jupyter import set_notebook_name
from jovian.utils.latest import check_update


def _initialize_jovian():
    """Initialize by setting the notebook name and checking for updates"""
    set_notebook_name()
    check_update()



import sys

import click

from jovian._version import __version__
from jovian.utils.clone import clone, pull
from jovian.utils.commit import commit_path
from jovian.utils.configure import configure, reset_config
from jovian.utils.extension import setup_extension
from jovian.utils.install import activate, install
from jovian.utils.logger import log
from jovian.utils.misc import is_py2
from jovian.utils.rcfile import set_notebook_slug
from jovian.utils.slack import add_slack


@click.group()
@click.version_option(version=__version__, prog_name="Jovian")
@click.pass_context
def main(ctx, log_level="info"):
    """Keep track of your Jupyter notebooks using Jovian.

    Use within your Jupyter notebooks:

        [1] import jovian

        [2] jovian.commit()

    Or try out a demo with:

        $ jovian clone aakashns/jovian-tutorial
    """

    pass


@main.command("help")
@click.pass_context
def help(ctx):  # no-cover
    """Print this help message."""

    # Pretend user typed 'jovian --help' instead of 'jovian help'.
    sys.argv[1] = "--help"
    main()


@main.command('version')
@click.pass_context
def main_version(ctx):  # no-cover
    """Print installed Jovian library version."""

    # Pretend user typed 'jovian --version' instead of 'jovian version'
    sys.argv[1] = "--version"
    main()


@main.command("configure")
@click.pass_context
def create_config(ctx):
    """Configure Jovian for Pro users."""

    configure()


@main.command("reset")
@click.pass_context
def reset(ctx):
    """Reset Jovian config."""
    reset_config()


@main.command("install", short_help="Install packages from environment file.")
@click.option('-n', '--name', 'name', help="Name of conda env")
@click.pass_context
def install_env(ctx, name=None):
    """Install packages from environment file:

        $ jovian install

    or, to specify environment name:

        $ jovian install -n my_env

    """

    if not name:
        install()
    else:
        install(env_name=name)


@main.command("activate")
@click.pass_context
def activate_env(ctx):
    """Activate conda environment from environment file."""

    activate()


@main.command("clone", short_help="Clone a notebook hosted on Jovian")
@click.argument('notebook', metavar="<username/title>")
@click.option('-v', '--version', 'version', help="Version number")
@click.option('--no-outputs', 'no_outputs', is_flag=True, default=False, help="Exclude output files")
@click.option('--overwrite', 'overwrite', is_flag=True, help="Overwrite existing project")
@click.pass_context
def exec_clone(ctx, notebook, version, no_outputs, overwrite):
    """Clone a notebook hosted on Jovian:

        $ jovian clone aakashns/jovian-tutorial

    Or clone a specific version of notebook:

        $ jovian clone aakashns/jovian-tutorial -v 10
    """
    clone(slug=notebook, version=version, include_outputs=not no_outputs, overwrite=overwrite)


@main.command("pull", short_help="Fetch new version of notebook hosted Jovian.")
@click.option('-n', '--notebook', 'notebook', help="Notebook project (format: username/title)")
@click.option('-v', '--version', 'version', help="Version number")
@click.pass_context
def exec_pull(ctx, notebook, version):
    """Fetch the new version of notebook hosted on Jovian(inside a cloned directory):

        $ jovian pull 

    Or fetch a specific version of a specific notebook:

        $ jovian pull -n aakashns/jovian-tutorial -v 10
    """

    pull(slug=notebook, version=version)


@main.command("set-project", short_help="Associate a notebook (filename.ipynb) to a Jovian project (username.title)")
@click.argument('notebook')
@click.argument('project')
@click.pass_context
def set_project(ctx,  notebook, project):
    """Associate notebook (filename.ipynb) to Jovian project (username/title)

        $ jovian set-project my_notebook.ipynb danb/keras-example

    This will create or update the .jovianrc file in the current directory to ensure that commits
    inside the Jupyter notebook "my_notebook.ipynb" add new versions to the project danb/keras-example
    """
    set_notebook_slug(filename=notebook, slug=project)


@main.command("commit", short_help="Create a new notebook on Jovian")
@click.argument('notebook')
@click.pass_context
def exec_commit(ctx, notebook):
    """Create a new notebook on Jovian

        $ jovian commit my_notebook.ipynb
    """
    if is_py2():
        log("Committing is not supported for Python 2.x. Please install and run Jovian from Python 3.5 and above.",
            warn=True)
    commit_path(path=notebook, environment=None, is_cli=True)


@main.command("add-slack")
@click.pass_context
def exec_add_slack(ctx):
    """Connect slack to get updates."""

    add_slack()


@main.command("enable-extension")
@click.pass_context
def extension_enable(ctx):
    """Enable Jovian’s Jupyter notebook extension."""

    setup_extension(enable=True)


@main.command("disable-extension")
@click.pass_context
def extension_disable(ctx):
    """Disable Jovian’s Jupyter notebook extension."""

    setup_extension(enable=False)


if __name__ == '__main__':
    main()



from torch import Tensor
from fastai.basic_train import Learner
from fastai.callback import Callback

from jovian.utils.records import log_hyperparams, log_metrics, reset
from jovian.utils.logger import log


class JovianFastaiCallback(Callback):
    """Fastai callback to automatically log hyperparameters and metrics.

    Args:
        learn (Learner): A learner object reference of your current model.

        arch_name (string): A name for the model you're training. 

    Example
        .. code-block::

            from jovian.callbacks.fastai import JovianFastaiCallback

            jvn_cb = JovianFastaiCallback(learn, 'res18')
            learn.fit_one_cycle(5, callbacks = jvn_cb)

    .. admonition:: Tutorial

        Visit `this`_ for a detailed example on using the fastai callback, also visit the *Records* tab
        to see all the logs of that notebook logged by the callback.
    .. _this: https://jovian.ml/PrajwalPrashanth/7f16274fc3224d829941bc2553ef6061?utm_source=docs
    """

    def __init__(self, learn: Learner, arch_name=None, reset_tracking=True):
        self.learn = learn
        self.arch_name = arch_name
        self.met_names = ['epoch', 'train_loss']
        # existence of validation dataset
        self.valid_set = self.learn.data.valid_dl.items.size > 0
        self.reset_tracking = reset_tracking
        if self.valid_set:
            self.met_names.append('valid_loss')

    def on_train_begin(self, n_epochs: int, metrics_names: list, **ka):
        if self.reset_tracking:
            reset('hyperparams', 'metrics')
        hyp_dict = {
            'epochs': n_epochs,
            'batch_size': self.learn.data.batch_size,
            'loss_func': str(self.learn.loss_func.func),
            'opt_func': str(self.learn.opt_func.func).split("'")[1],
            'weight_decay': self.learn.wd,
            'learning_rate': str(self.learn.opt.lr)
        }
        if self.arch_name:
            hyp_dict['arch_name'] = self.arch_name
        log_hyperparams(hyp_dict)

        if self.valid_set:
            self.met_names.extend(metrics_names)

    def on_epoch_end(self, epoch: int, smooth_loss: Tensor, last_metrics: list, **ka):
        met_values = [epoch,
                      smooth_loss.item()]  # smoothened avg. train loss for the epoch

        if self.valid_set:
            # last_metrics is a list with first elem as valid_loss followed by all
            # the metrics of the learner
            met_values.extend([str(last_metrics[0])] + [i.item()
                                                        for i in last_metrics[1:]])
        log_metrics(dict(zip(self.met_names, met_values)))

    def on_train_end(self, **ka):  # no-cover
        if not self.valid_set:
            log('Metrics apart from train_loss are not calculated in fastai without a validation dataset')



from keras.backend import get_value
from keras.callbacks import Callback
import json
import numpy

from jovian.utils.records import log_metrics, log_hyperparams, reset
from jovian.utils.slack import notify


class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        else:  # no-cover
            return super(Encoder, self).default(obj)


class JovianKerasCallback(Callback):
    """Keras Callback to log hyperparameters and metrics during model training.

    Args:
        reset_tracking (string, optional): Will clear previously tracked hyperparameters & metrics, and start a fresh recording. Defaults to True.
        arch_name (string, optional): A name for the model you’re training.
        every_epoch (bool, optional): Whether to record losses & metrics for every epoch or just the final loss & metric. Defaults to False.
        notify (bool, optional): Whether to send notification on slack when the training ends. Defaults to False.

    Example
        .. code-block::

            from jovian.callbacks.keras import JovianKerasCallback

            # To record logs of every epoch and to notify on slack
            jvn_cb = JovianKerasCallback(arch_name='resnet18', every_epoch=True, notify=True)
            model.fit(x_train, y_train, ...., callbacks=[jvn_cb])

    .. admonition:: Tutorial

        Visit `this`_ for a detailed example on using the fastai callback, also visit the *Records* tab
        to see all the logs of that notebook logged by the callback.
    .. _this: https://jovian.ml/PrajwalPrashanth/34fd4e72905e460db2d16aafab285537?utm_source=docs

    """

    def __init__(self, reset_tracking=True, arch_name='', every_epoch=False, notify=False):
        self.arch_name = arch_name
        self.every_epoch = every_epoch
        self.reset_tracking = reset_tracking
        self.notify_slack = notify

    def on_train_begin(self, logs=None):
        # Reset state if required
        if self.reset_tracking:
            reset('hyperparams', 'metrics')

        hyp_dict = {
            'epochs': self.params['epochs'],
            'batch_size': self.params['batch_size'],
            'loss_func': self.model.loss,
            'opt': str(self.model.optimizer.__class__).split("'")[1].split('.')[-1],
            'wt_decay': self.model.optimizer.initial_decay,
            'lr': str(get_value(self.model.optimizer.lr))
        }
        if self.arch_name:
            hyp_dict['arch'] = self.arch_name
        log_hyperparams(hyp_dict, verbose=False)
        self.hyperparams = hyp_dict

    def on_epoch_end(self, epoch, logs):
        if self.every_epoch:
            met_dict = {'epoch': epoch}
            for key, value in logs.items():
                logs[key] = round(value, 4)
            met_dict.update(logs)
            log_metrics(met_dict, verbose=False)

        elif epoch == self.params['epochs'] - 1:
            met_dict = {}
            for key, value in logs.items():
                logs[key] = round(value, 4)
            met_dict.update(logs)
            log_metrics(met_dict, verbose=False)

            if self.notify_slack:
                result = {
                    'message': 'Training complete after ' + str(self.params['epochs']) + ' epochs.',
                    'metrics': met_dict
                }
                if self.hyperparams:
                    result['hyperparams'] = self.hyperparams
                notify(json.dumps(result, indent=2, cls=Encoder), verbose=False, safe=True)



VALID_EXT = ["py"]


def get_script_filename():
    try:
        from __main__ import __file__
        return __file__
    except:
        return None


def in_script():
    filename = get_script_filename()
    if filename is not None:
        ext = filename.split(".")[-1]
        return ext in VALID_EXT

    return False



"""Utilies to manage the .jovianrc file"""
import os
import json
from jovian.utils.constants import RC_FILENAME

_current_slug = None


def rcfile_exists():
    """Check if .jovianrc exists"""
    return os.path.exists(RC_FILENAME)


def save_rcdata(data=None):
    """Save given data (or empty) to file"""
    if data is None:
        data = {'notebooks': {}}
    with open(RC_FILENAME, 'w') as f:
        json.dump(data, f, indent=2)


def get_rcdata():
    """Read .jovianrc file as a dict"""
    if not rcfile_exists():
        save_rcdata()
    with open(RC_FILENAME, 'r') as f:
        data = json.load(f)
    return data


def get_notebook_slug(filename, cache_result=True):
    """Get the gist slug for a notebook filename"""
    global _current_slug
    data = get_rcdata()
    if (filename in data.get('notebooks', {})):
        slug = data['notebooks'][filename]['slug']
        if slug and cache_result:
            _current_slug = slug
        return slug


def get_cached_slug():
    """Get the cached notebook slug"""
    global _current_slug
    return _current_slug


def set_notebook_slug(filename, slug, cache_result=True):
    """Set the gist slug for a notebook filename"""
    global _current_slug
    data = get_rcdata()
    data['notebooks'][filename] = {'slug': slug}
    save_rcdata(data)
    if cache_result:
        _current_slug = slug
    return slug


def reset_notebook_slug():
    """Reset cached notebook slug"""
    global _current_slug
    _current_slug = None


def make_rcdata(filename, slug):
    """Make a JSON string for an individual file"""
    data = {
        "notebooks": {
            filename: {
                "slug": slug
            }
        }
    }
    return json.dumps(data)



import os

import click
import yaml

from jovian.utils.constants import LINUX, MACOS, WINDOWS
from jovian.utils.logger import log
from jovian.utils.misc import get_platform

blacklist = [
    'conda',
    'conda-env',
    'attrs'
]


class Loader(yaml.SafeLoader):
    def __init__(self, stream):
        self._root = os.path.split(stream.name)[0]
        super(Loader, self).__init__(stream)


def get_environment_dict(env_fname):
    """Extract the environment as a python dict"""
    try:
        with open(env_fname, 'r') as stream:
            environment = yaml.load(stream=stream, Loader=Loader)
            return environment
    except yaml.YAMLError as e:
        print("Exception: {} while reading file {}".format(e, env_fname))
        return {}


def dump_environment_to_yaml_file(env_fname, environment):
    """Dump environment into yaml file"""
    try:
        with open(env_fname, 'w') as stream:
            yaml.safe_dump(data=environment, stream=stream, default_style=None,
                           default_flow_style=False)

    except yaml.YAMLError as e:  # no-cover
        print("Exception: {} while accessing file {}".format(e, env_fname))


def write_env_name(env_name, env_fname):
    """Overwrite the environment name into the file"""
    environment = get_environment_dict(env_fname=env_fname)
    if environment and (not environment.get('name') or environment.get('name') != env_name):
        environment['name'] = env_name
        dump_environment_to_yaml_file(
            env_fname=env_fname, environment=environment)


ENV_NAME_MSG = "Please provide a name for the conda environment"


def request_env_name(env_name, env_fname):
    """Request the user to provide a name for the environment"""
    if env_name is None:
        env_name = extract_env_name(env_fname)
        # Make sure we're not overwriting the base environment
        # Construct the help message with default value
        # Prompt the user for input
        if env_name is None or env_name == 'base':
            user_input = click.prompt(ENV_NAME_MSG, default='base', show_default=False)
        else:
            user_input = click.prompt(ENV_NAME_MSG, default=env_name, show_default=True)

        print('')
        # Sanitize the input
        user_input = user_input.strip()
        # Set the final env name
        if user_input:
            env_name = user_input
        if env_name:
            # Write the chosen name to file
            write_env_name(env_name, env_fname)
        else:
            env_name = 'base'

    return env_name


MISSING_MSG = ("WARNING: Some packages listed in the environment definition file could" +
               " not be installed, possibly because the environment was recorded on a different" +
               " operating system. As a result, you have to install some packages manually using " +
               "'conda install <package_name>' if you face errors while executing the code.\n")


def check_error(error_str, packages=None):
    """Check if the error output contains ResolvePackageNotFound or UnsatisfiableError"""
    if not packages:
        packages = []
    error_lines = error_str.split('\n')
    error = None
    pkgs = []
    for line in error_lines:
        if 'ResolvePackageNotFound:' in line:
            error = 'unresolved'
        elif 'UnsatisfiableError:' in line:
            error = 'unsatisfiable'
            log(MISSING_MSG)
        if error:
            pkg = extract_package_from_line(line, packages)
            if pkg and pkg not in pkgs:
                pkgs.append(pkg)
    return error, pkgs


def check_pip_failed(error_str):
    """Check if the error output contains Pip failed message"""
    error_lines = error_str.split('\n')
    for line in error_lines:
        if 'Pip failed' in line:
            return True
    return False


def identify_env_file(env_fname, folder_prefix=""):
    """Find the right conda environment file through trial and errors"""

    if env_fname is None:
        # Look for platform specific environment files (prefer current)
        platforms = [get_platform()] + ["", LINUX, WINDOWS, MACOS]
        for platform in platforms:
            if platform == "":
                expected_fname = os.path.join(folder_prefix, 'environment.yml')
            else:
                expected_fname = os.path.join(folder_prefix, 'environment-' + platform + '.yml')

            if os.path.exists(expected_fname):
                env_fname = expected_fname
                break

    return env_fname


def extract_package_from_line(line, packages):
    """Extract the name of a package from an error line"""
    for p in packages:
        if p in line.strip():
            return p
    """if exact version not found in string, look just for package name"""
    for p in packages:
        if '=' in p and p.split('=')[0] in line.strip():
            return p
    return None


def extract_env_name(env_fname):
    """Extract the name of the environment from the env file"""
    environment = get_environment_dict(env_fname=env_fname)
    name = environment.get('name') if environment else None
    return name


def extract_env_packages(env_fname):
    """Extract the (iterable) packages of the environment from the env file
                (including the pip packages in the same list)           """
    environment = get_environment_dict(env_fname=env_fname)
    if not environment:
        return []
    dependencies = environment.get('dependencies')
    return serialize_packages(dependencies=dependencies)


def extract_pip_packages(env_fname):
    """Extract the pip packages of the environment from the env file"""
    environment = get_environment_dict(env_fname=env_fname)
    if not environment:
        return []
    dependencies = environment.get('dependencies')
    if len(dependencies) > 0:
        dependencies[:] = (d for d in dependencies if isinstance(
            d, dict) and d.get('pip'))
    return serialize_packages(dependencies=dependencies)


def serialize_packages(dependencies):
    serialized_dependencies = []
    for i, dependency in enumerate(dependencies):
        if isinstance(dependency, str):
            serialized_dependencies.append(dependency)
        elif isinstance(dependency, dict) and dependency.get('pip') and len(dependency['pip']) > 0:
            for d in dependency['pip']:
                serialized_dependencies.append(d)
    return serialized_dependencies


def remove_packages(dependencies, pkgs):
    new_dependencies = []
    for i, dependency in enumerate(dependencies):
        if isinstance(dependency, str):
            if dependency not in pkgs and dependency not in blacklist:
                new_dependencies.append(dependency)
        elif isinstance(dependency, dict) and dependency.get('pip') and len(dependency['pip']) > 0:
            new_pip_dependencies = remove_packages(
                dependencies=dependency['pip'], pkgs=pkgs)
            new_dependencies.append({
                'pip': new_pip_dependencies
            })
    return new_dependencies


def sanitize_envfile(env_fname, pkgs):
    """Remove the given packages from the environment file"""
    environment = get_environment_dict(env_fname=env_fname)
    dependencies = environment.get('dependencies')
    dependencies = remove_packages(dependencies=dependencies, pkgs=pkgs)
    environment['dependencies'] = dependencies

    dump_environment_to_yaml_file(env_fname=env_fname, environment=environment)



import sys

from jovian.utils.api import get_current_user, _h
from jovian.utils.error import ApiError
from jovian.utils.logger import log
from jovian.utils.request import post, pretty
from jovian.utils.shared import _u


_colab_file_id = None


def in_colab():
    return 'google.colab' in sys.modules


def set_colab_file_id(file_id):
    global _colab_file_id
    _colab_file_id = file_id


def get_colab_file_id():
    global _colab_file_id
    return _colab_file_id


def perform_colab_commit(project, privacy):
    # file_id, project, privacy api key
    file_id = get_colab_file_id()
    if file_id is None:
        log("Colab File Id is not provided", error=True)

    # /gist/colab-commit data = {file_id, project}, return status
    if '/' not in project:
        project = get_current_user()['username'] + '/' + project

    data = {'project': project, 'file_id': file_id, 'visibility': privacy}

    if privacy == 'auto':
        data['public'] = True
    elif privacy == 'secret' or privacy == 'private':
        data['public'] = False

    auth_headers = _h()

    log("Uploading colab notebook to Jovian...")
    res = post(url=_u('/gist/colab-commit'),
               data=data,
               headers=auth_headers)

    if res.status_code == 200:
        return res.json()['data']
    raise ApiError('Colab commit failed: ' + pretty(res))



"""Jupyter related utilities"""
from requests.compat import urljoin
import json
import os.path
import re
import requests
import time
from io import StringIO
import sys


def has_ipynb_shell():
    """Check if IPython shell is available"""
    try:
        from IPython import get_ipython
        cls = get_ipython().__class__.__name__
        return cls == 'ZMQInteractiveShell'
    except:
        return False


def in_notebook():
    """Check if this code is being executed in a notebook"""
    if not has_ipynb_shell():
        return False
    from ipykernel.kernelapp import IPKernelApp
    return IPKernelApp.initialized()


def get_notebook_server_path():
    try:  # Python 3
        from notebook.notebookapp import list_running_servers
    except ImportError:  # no-cover
        # Python 2
        import warnings
        from IPython.utils.shimmodule import ShimWarning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ShimWarning)
            from IPython.html.notebookapp import list_running_servers

    """Get the path of the notebook relative to the Jupyter server"""
    import ipykernel
    kernel_id = re.search('kernel-(.*).json',
                          ipykernel.connect.get_connection_file()).group(1)
    servers = list_running_servers()
    for ss in servers:
        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                params={'token': ss.get('token', '')})
        for nn in json.loads(response.text):
            if nn['kernel']['id'] == kernel_id:
                relative_path = nn['notebook']['path']
                return relative_path


def get_notebook_path_py():
    try:  # Python 3
        from notebook.notebookapp import list_running_servers
    except ImportError:  # no-cover
        # Python 2
        import warnings
        from IPython.utils.shimmodule import ShimWarning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ShimWarning)
            from IPython.html.notebookapp import list_running_servers

    import ipykernel
    """Get the full path of the jupyter notebook."""
    kernel_id = re.search('kernel-(.*).json',
                          ipykernel.connect.get_connection_file()).group(1)
    servers = list_running_servers()
    for ss in servers:
        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                params={'token': ss.get('token', '')})
        if response.status_code == 200:
            for nn in json.loads(response.text):
                if nn["kernel"] and nn['kernel']['id'] == kernel_id:
                    relative_path = nn['notebook']['path']
                    return os.path.join(ss['notebook_dir'], relative_path)


def get_notebook_path():
    nb_name = None
    # Try getting the notebook name using IPython API
    try:
        nb_name = get_notebook_path_py()
    except:
        pass

    # Try using Javascript instead
    if nb_name is None:
        saved_name = get_notebook_name_saved()
        if saved_name:
            nb_name = os.path.join(os.getcwd(), saved_name)

    return nb_name


def set_notebook_name():
    if in_notebook():
        from IPython import get_ipython
        get_ipython().run_cell_magic(
            'javascript', '',
            "if (window.IPython && IPython.notebook.kernel) IPython.notebook.kernel.execute('jovian.utils.jupyter.get_notebook_name_saved = lambda: \"' + IPython.notebook.notebook_name + '\"')")


def get_notebook_name_saved():
    return None


def get_notebook_name():
    """Return the name of the notebook"""
    nb_path = get_notebook_path()
    if nb_path:
        return os.path.basename(nb_path)
    if nb_path and not os.path.exists(nb_path):
        set_notebook_name()  # no-cover
    return None


def get_notebook_history():
    from IPython import get_ipython
    """Return full code history of notebook"""
    return get_ipython().magic('%history')


def save_notebook():
    from IPython import get_ipython
    """Save the current Jupyter notebook"""
    return get_ipython().run_cell_magic(
        'javascript',
        '',
        'window.require && require(["base/js/namespace"],function(Jupyter){Jupyter.notebook.save_checkpoint()})')
